"""
pyndl.ndl
---------

*pyndl.ndl* provides functions in order to train NDL models

"""
from collections import defaultdict, OrderedDict
import copy
import getpass
import os
from queue import Queue
import socket
import sys
import tempfile
import threading
import time
import warnings
import types

import cython
import pandas as pd
import numpy as np
import xarray as xr

from . import __version__ as pyndl_version
from . import count
from . import preprocess
from . import ndl_parallel
from . import io

# conditional import as openmp is only compiled for linux
if sys.platform.startswith('linux'):
    from . import ndl_openmp
elif sys.platform.startswith('win32'):
    pass
elif sys.platform.startswith('darwin'):
    pass


warnings.simplefilter('always', DeprecationWarning)


class WeightDict(defaultdict):
    # pylint: disable=missing-docstring

    """
    Subclass of defaultdict to represent outcome-cue weights.

    Notes
    -----
    Weight for each outcome-cue combination is 0 per default.

    """

    # pylint: disable=W0613
    def __init__(self, *args, **kwargs):
        super().__init__(lambda: defaultdict(float))

        self._attrs = OrderedDict()

        if 'attrs' in kwargs:
            self.attrs = kwargs['attrs']
        else:
            self.attrs = {}

    @property
    def attrs(self):
        return self._attrs

    @attrs.setter
    def attrs(self, attrs):
        self._attrs = OrderedDict(attrs)


def ndl(events, alpha, betas, lambda_=1.0, *,
        method='openmp', weights=None,
        number_of_threads=None, n_jobs=8, len_sublists=None, n_outcomes_per_job=10,
        remove_duplicates=None,
        verbose=False, temporary_directory=None,
        events_per_temporary_file=10000000):
    """
    Calculate the weights for all_outcomes over all events in event_file
    given by the files path.

    This is a parallel python implementation using numpy, multithreading and
    the binary format defined in preprocess.py.

    Parameters
    ----------
    events : generator or str
        generates cues, outcomes pairs or the path to the event file
    alpha : float
        saliency of all cues
    betas : (float, float)
        one value for successful prediction (reward) one for punishment
    lambda\\_ : float

    method : {'openmp', 'threading'}
    weights : None or xarray.DataArray
        the xarray.DataArray needs to have the dimensions 'cues' and 'outcomes'
    n_jobs : int
        a integer giving the number of threads in which the job should
        executed
    n_outcomes_per_job : int
        a integer giving the length of sublists generated from all outcomes
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    verbose : bool
        print some output if True.
    temporary_directory : str
        path to directory to use for storing temporary files created;
        if none is provided, the operating system's default will
        be used (/tmp on unix)
    events_per_temporary_file: int
        Number of events in each temporary binary file. Has to be larger than 1

    Returns
    -------
    weights : xarray.DataArray
        with dimensions 'outcomes' and 'cues'. You can lookup the weights
        between a cue and an outcome with ``weights.loc[{'outcomes': outcome,
        'cues': cue}]`` or ``weights.loc[outcome].loc[cue]``.

    """

    # Create temporary file if events is a generator
    if isinstance(events, types.GeneratorType):
        file_path = tempfile.NamedTemporaryFile().name
        io.events_to_file(events, file_path)
        events = file_path
        del file_path

    if number_of_threads is not None:
        warnings.warn("Parameter `number_of_threads` is renamed to `n_jobs`. The old name "
                      "will stop working with v0.9.0.",
                      DeprecationWarning, stacklevel=2)
        n_jobs = number_of_threads
    if len_sublists is not None:
        warnings.warn("Parameter `len_sublists` is renamed to `n_outcomes_per_job`. The old name "
                      "will stop working with v0.9.0.",
                      DeprecationWarning, stacklevel=2)
        n_outcomes_per_job = len_sublists

    if not (remove_duplicates is None or isinstance(remove_duplicates, bool)):
        raise ValueError("remove_duplicates must be None, True or False")
    if not isinstance(events, str):
        raise ValueError("'events' need to be the path to a gzipped event file not {}".format(type(events)))

    weights_ini = weights
    wall_time_start = time.perf_counter()
    cpu_time_start = time.process_time()

    # preprocessing
    n_events, cues, outcomes = count.cues_outcomes(events,
                                                   n_jobs=n_jobs,
                                                   verbose=verbose)
    cues = list(cues.keys())
    outcomes = list(outcomes.keys())
    cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues)))
    outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(outcomes)))

    all_outcome_indices = [outcome_map[outcome] for outcome in outcomes]

    shape = (len(outcome_map), len(cue_map))

    # initialize weights
    if weights is None:
        weights = np.ascontiguousarray(np.zeros(shape, dtype=np.float64, order='C'))
    elif isinstance(weights, xr.DataArray):
        old_cues = weights.coords["cues"].values.tolist()
        new_cues = list(set(cues) - set(old_cues))
        old_outcomes = weights.coords["outcomes"].values.tolist()
        new_outcomes = list(set(outcomes) - set(old_outcomes))

        cues = old_cues + new_cues
        outcomes = old_outcomes + new_outcomes

        cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues)))
        outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(outcomes)))

        all_outcome_indices = [outcome_map[outcome] for outcome in outcomes]

        weights_tmp = np.concatenate((weights.values,
                                      np.zeros((len(new_outcomes), len(old_cues)),
                                               dtype=np.float64, order='C')),
                                     axis=0)
        weights_tmp = np.concatenate((weights_tmp,
                                      np.zeros((len(outcomes), len(new_cues)),
                                               dtype=np.float64, order='C')),
                                     axis=1)

        weights = np.ascontiguousarray(weights_tmp)

        del weights_tmp, old_cues, new_cues, old_outcomes, new_outcomes
    else:
        raise ValueError('weights need to be None or xarray.DataArray with method=%s' % method)

    if any(length > 4294967295 for length in weights.shape):
        raise ValueError("Neither number of cues nor outcomes shall exceed 4294967295 "
                         "for now. See https://github.com/quantling/pyndl/issues/169")

    beta1, beta2 = betas

    with tempfile.TemporaryDirectory(prefix="pyndl", dir=temporary_directory) as binary_path:
        number_events = preprocess.create_binary_event_files(events, binary_path, cue_map,
                                                             outcome_map, overwrite=True,
                                                             n_jobs=n_jobs,
                                                             events_per_file=events_per_temporary_file,
                                                             remove_duplicates=remove_duplicates,
                                                             verbose=verbose)
        assert n_events == number_events, (str(n_events) + ' ' + str(number_events))
        binary_files = [os.path.join(binary_path, binary_file)
                        for binary_file in os.listdir(binary_path)
                        if os.path.isfile(os.path.join(binary_path, binary_file))]
        # sort binary files as they were created
        binary_files.sort(key=lambda filename: int(os.path.basename(filename)[9:-4]))
        if verbose:
            print('start learning...')
        # learning
        if not weights.data.c_contiguous:
            raise ValueError('weights has to be c_contiguous')
        if method == 'openmp':
            if not sys.platform.startswith('linux'):
                raise NotImplementedError("OpenMP is linux only at the moment."
                                          "Use method='threading' instead.")
            ndl_openmp.learn_inplace_binary_to_binary(binary_files,
                                                      alpha,
                                                      beta1,
                                                      beta2,
                                                      lambda_,
                                                      weights,
                                                      np.array(all_outcome_indices,
                                                               dtype=np.uint32),
                                                      n_outcomes_per_job,
                                                      n_jobs)
        elif method == 'threading':
            part_lists = slice_list(all_outcome_indices, n_outcomes_per_job)

            working_queue = Queue(len(part_lists))
            threads = []
            queue_lock = threading.Lock()

            def worker():
                while True:
                    with queue_lock:
                        if working_queue.empty():
                            break
                        data = working_queue.get()
                    ndl_parallel.learn_inplace_binary_to_binary(binary_files,
                                                                alpha,
                                                                beta1,
                                                                beta2,
                                                                lambda_,
                                                                weights,
                                                                data)

            with queue_lock:
                for partlist in part_lists:
                    working_queue.put(np.array(partlist, dtype=np.uint32))

            for _ in range(n_jobs):
                thread = threading.Thread(target=worker)
                thread.start()
                threads.append(thread)

            for thread in threads:
                thread.join()
        else:
            raise ValueError('method needs to be either "threading" or "openmp"')

    cpu_time_stop = time.process_time()
    wall_time_stop = time.perf_counter()
    cpu_time = cpu_time_stop - cpu_time_start
    wall_time = wall_time_stop - wall_time_start

    if weights_ini is not None:
        attrs_to_be_updated = weights_ini.attrs
    else:
        attrs_to_be_updated = None

    attrs = _attributes(events, number_events, alpha, betas, lambda_, cpu_time, wall_time,
                        __name__ + "." + ndl.__name__, method=method, attrs=attrs_to_be_updated)

    # post-processing
    weights = xr.DataArray(weights, [('outcomes', outcomes), ('cues', cues)],
                           attrs=attrs)
    return weights


def _attributes(event_path, number_events, alpha, betas, lambda_, cpu_time,
                wall_time, function, method=None, attrs=None):
    if not isinstance(alpha, (float, int)):
        alpha_str = 'varying'
    else:
        alpha_str = str(alpha)

    width = max([len(ss) for ss in (event_path,
                                    str(number_events),
                                    str(alpha),
                                    str(betas),
                                    str(lambda_),
                                    function,
                                    str(method),
                                    socket.gethostname(),
                                    getpass.getuser())])
    width = max(19, width)

    def _format(value):
        return '{0: <{width}}'.format(value, width=width)

    new_attrs = {'date': _format(time.strftime("%Y-%m-%d %H:%M:%S")),
                 'event_path': _format(event_path),
                 'number_events': _format(number_events),
                 'alpha': _format(alpha_str),
                 'betas': _format(str(betas)),
                 'lambda': _format(str(lambda_)),
                 'function': _format(function),
                 'method': _format(str(method)),
                 'cpu_time': _format(str(cpu_time)),
                 'wall_time': _format(str(wall_time)),
                 'hostname': _format(socket.gethostname()),
                 'username': _format(getpass.getuser()),
                 'pyndl': _format(pyndl_version),
                 'numpy': _format(np.__version__),
                 'pandas': _format(pd.__version__),
                 'xarray': _format(xr.__version__),
                 'cython': _format(cython.__version__)}

    if attrs is not None:
        for key in set(attrs.keys()) | set(new_attrs.keys()):
            if key in attrs:
                old_val = attrs[key]
            else:
                old_val = ''
            if key in new_attrs:
                new_val = new_attrs[key]
            else:
                new_val = ''
            new_attrs[key] = old_val + ' | ' + new_val
    return new_attrs


def dict_ndl(events, alphas, betas, lambda_=1.0, *,
             weights=None, inplace=False, remove_duplicates=None,
             make_data_array=False, verbose=False):
    """
    Calculate the weights for all_outcomes over all events in event_file.

    This is a pure python implementation using dicts.

    Notes
    -----
    The metadata will only be stored when `make_data_array` is True and then
    `dict_ndl` cannot be used to continue learning. At the moment there is no
    proper way to automatically store the meta data into the default dict.

    Parameters
    ----------
    events : generator or str
        generates cues, outcomes pairs or the path to the event file
    alphas : dict or float
        a (default)dict having cues as keys and a value below 1 as value
    betas : (float, float)
        one value for successful prediction (reward) one for punishment
    lambda\\_ : float
    weights : dict of dicts or xarray.DataArray or None
        initial weights
    inplace: {True, False}
        if True calculates the weightmatrix inplace
        if False creates a new weightmatrix to learn on
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    make_data_array : {False, True}
        if True makes a xarray.DataArray out of the dict of dicts.
    verbose : bool
        print some output if True.

    Returns
    -------
    weights : dict of dicts of floats
        the first dict has outcomes as keys and dicts as values
        the second dict has cues as keys and weights as values
        weights[outcome][cue] gives the weight between outcome and cue.

    or

    weights : xarray.DataArray
        with dimensions 'outcomes' and 'cues'. You can lookup the weights
        between a cue and an outcome with ``weights.loc[{'outcomes': outcome,
        'cues': cue}]`` or ``weights.loc[outcome].loc[cue]``.

    """

    if not isinstance(make_data_array, bool):
        raise ValueError("make_data_array must be True or False")

    if not (remove_duplicates is None or isinstance(remove_duplicates, bool)):
        raise ValueError("remove_duplicates must be None, True or False")

    wall_time_start = time.perf_counter()
    cpu_time_start = time.process_time()
    if isinstance(events, str):
        event_path = events
    else:
        event_path = ""
    attrs_to_update = None

    # weights can be seen as an infinite outcome by cue matrix
    # weights[outcome][cue]
    if weights is None:
        weights = WeightDict()
    elif isinstance(weights, WeightDict):
        attrs_to_update = weights.attrs
    elif isinstance(weights, xr.DataArray):
        weights_ini = weights
        attrs_to_update = weights_ini.attrs
        coords = weights_ini.coords
        weights = WeightDict()
        for outcome_index, outcome in enumerate(coords['outcomes'].values):
            for cue_index, cue in enumerate(coords['cues'].values):
                weights[outcome][cue] = weights_ini.item((outcome_index, cue_index))
    elif not isinstance(weights, defaultdict):
        raise ValueError('weights needs to be either defaultdict or None')

    if not inplace:
        weights = copy.deepcopy(weights)

    beta1, beta2 = betas
    all_outcomes = set(weights.keys())

    if isinstance(events, str):
        events = io.events_from_file(events)
    if isinstance(alphas, float):
        alpha = alphas
        alphas = defaultdict(lambda: alpha)
    number_events = 0

    for cues, outcomes in events:
        number_events += 1
        if verbose and number_events % 1000:
            print('.', end='')
            sys.stdout.flush()
        if remove_duplicates is None:
            if (len(cues) != len(set(cues)) or
                    len(outcomes) != len(set(outcomes))):
                raise ValueError('cues or outcomes needs to be unique: cues '
                                 '"%s"; outcomes "%s"; use '
                                 'remove_duplicates=True' %
                                 (' '.join(cues), ' '.join(outcomes)))
        elif remove_duplicates:
            cues = set(cues)
            outcomes = set(outcomes)
        else:
            pass

        all_outcomes.update(outcomes)
        for outcome in all_outcomes:
            association_strength = sum(weights[outcome][cue] for cue in cues)
            if outcome in outcomes:
                update = beta1 * (lambda_ - association_strength)
            else:
                update = beta2 * (0 - association_strength)
            for cue in cues:
                weights[outcome][cue] += alphas[cue] * update

    cpu_time_stop = time.process_time()
    wall_time_stop = time.perf_counter()
    cpu_time = cpu_time_stop - cpu_time_start
    wall_time = wall_time_stop - wall_time_start
    attrs = _attributes(event_path, number_events, alphas, betas, lambda_, cpu_time, wall_time,
                        __name__ + "." + dict_ndl.__name__, attrs=attrs_to_update)

    if make_data_array:
        weights = data_array(weights, attrs=attrs)
    else:
        weights.attrs = attrs

    return weights


def data_array(weights, *, attrs=None):
    """
    Calculate the weights for all_outcomes over all events in event_file.

    Parameters
    ----------
    weights : dict of dicts of floats or WeightDict
        the first dict has outcomes as keys and dicts as values
        the second dict has cues as keys and weights as values
        weights[outcome][cue] gives the weight between outcome and cue.
        If a dict of dicts is given, attrs is required. If a WeightDict is
        given, attrs is optional
    attrs : dict
        A dictionary of attributes

    Returns
    -------
    weights : xarray.DataArray
        with dimensions 'outcomes' and 'cues'. You can lookup the weights
        between a cue and an outcome with ``weights.loc[{'outcomes': outcome,
        'cues': cue}]`` or ``weights.loc[outcome].loc[cue]``.
    """

    if isinstance(weights, xr.DataArray) and weights.dims == ('outcomes', 'cues'):
        return weights

    if attrs is None:
        try:
            attrs = weights.attrs
        except AttributeError:
            raise AttributeError("weights does not have attributes and no attrs "
                                 "argument is given.")

    outcomes = list(weights.keys())
    cues = set()
    for outcome in outcomes:
        cues.update(set(weights[outcome].keys()))

    cues = list(cues)

    weights_dict = weights
    shape = (len(outcomes), len(cues))
    weights = xr.DataArray(np.zeros(shape), attrs=attrs,
                           coords={'outcomes': outcomes, 'cues': cues},
                           dims=('outcomes', 'cues'))

    for outcome in outcomes:
        for cue in cues:
            weights.loc[{"outcomes": outcome, "cues": cue}] = weights_dict[outcome][cue]

    return weights


def slice_list(list_, len_sublists):
    r"""
    Slices a list in sublists with the length len_sublists.

    Parameters
    ----------
    list\_ : list
         list which should be sliced in sublists
    len_sublists : int
         integer which determines the length of the sublists

    Returns
    -------
    seq_list : list of lists
        a list of sublists with the length len_sublists

    """
    if len_sublists < 1:
        raise ValueError("'len_sublists' must be larger then one")
    assert len(list_) == len(set(list_))
    ii = 0
    seq_list = list()
    while ii < len(list_):
        seq_list.append(list_[ii:ii + len_sublists])
        ii = ii + len_sublists

    return seq_list



#!/usr/bin/env python3

# pylint: disable=C0111

import pytest
import numpy as np

from pyndl import correlation


@pytest.mark.nolinux
def test_correlation():
    np.random.seed(20190507)

    n_vec_dims = 40
    n_outcomes = 50
    n_cues = 20
    n_events = 120

    semantics = np.asfortranarray(np.random.random((n_vec_dims, n_outcomes)))

    weights = np.random.random((n_vec_dims, n_cues))
    events = np.random.random((n_cues, n_events))

    # n_vec_dims x n_events
    activations = np.asfortranarray(weights @ events)

    corr1 = correlation._reference_correlation(semantics, activations, verbose=True)
    corr2 = correlation.correlation(semantics, activations, verbose=True)

    assert np.allclose(corr1, corr2)


# TODO test for inputs that are NaN or have zero std.

# @jit(nopython=True)

# %time X2 = reference_corr(semantics, activations)
# CPU times: user 1min 11s, sys: 0 ns, total: 1min 11s
# Wall time: 1min 11s

# without numba jit
# %time X = manual_corr(semantics, activations)
# time needed for stds and means:  0.5793533325195312
# time needed for correlations:  19.464678525924683
# CPU times: user 20 s, sys: 0 ns, total: 20 s
# Wall time: 20 s

# np.allclose(X, X2)

# events 120
# time needed for correlations:  57.368293046951294
# time needed for stds and means:  0.3004155158996582
# time needed for correlations:  0.8774964809417725


# cython - events 1200

# %time X = manual_corr(semantics, activations)
# time needed for stds and means:  0.19872736930847168
# time needed for correlations:  4.957022190093994
# CPU times: user 29.9 s, sys: 32 ms, total: 29.9 s
# Wall time: 5.16 s

# %time X2 = reference_corr(semantics, activations)
# CPU times: user 5min 32s, sys: 20 ms, total: 5min 32s
# Wall time: 5min 32s

# cython - events 12000

# %time X = manual_corr(semantics, activations)
# time needed for stds and means:  0.556868314743042
# time needed for correlations:  56.93205285072327
# CPU times: user 5min 41s, sys: 392 ms, total: 5min 42s
# Wall time: 57.5 s

# import imp
# imp.reload(low_level_corr)


# python setup.py build_ext --inplace
# %run



#!/usr/bin/env python3

# pylint: disable=C0111

import time
import gc
import os
from collections import defaultdict

import numpy as np
import xarray as xr

import pytest

from pyndl import ndl
from pyndl.activation import activation


TEST_ROOT = os.path.join(os.path.pardir, os.path.dirname(__file__))
FILE_PATH_SIMPLE = os.path.join(TEST_ROOT, "resources/event_file_simple.tab.gz")
FILE_PATH_MULTIPLE_CUES = os.path.join(TEST_ROOT, "resources/event_file_multiple_cues.tab.gz")

LAMBDA_ = 1.0
ALPHA = 0.1
BETAS = (0.1, 0.1)


def test_exceptions():
    with pytest.raises(ValueError) as e_info:
        weights = ndl.dict_ndl(FILE_PATH_SIMPLE, ALPHA, BETAS, remove_duplicates=None)
        activation(FILE_PATH_MULTIPLE_CUES, weights)
        assert e_info == 'cues or outcomes needs to be unique: cues "a a"; outcomes "A"; use remove_duplicates=True'

    with pytest.raises(ValueError) as e_info:
        activation(FILE_PATH_MULTIPLE_CUES, "magic")
        assert e_info == "Weights other than xarray.DataArray or dicts are not supported."


def test_activation_matrix():
    weights = xr.DataArray(np.array([[0, 1, 0], [1, 0, 0]]),
                           coords={
                               'outcomes': ['o1', 'o2'],
                               'cues': ['c1', 'c2', 'c3']
                           },
                           dims=('outcomes', 'cues'))

    events = [(['c1', 'c2', 'c3'], []),
              (['c1', 'c3'], []),
              (['c2'], []),
              (['c1', 'c1'], [])]
    reference_activations = np.array([[1, 0, 1, 0], [1, 1, 0, 1]])

    with pytest.raises(ValueError):
        activations = activation(events, weights, n_jobs=1)

    activations = activation(events, weights, n_jobs=1, remove_duplicates=True)
    activations_mp = activation(events, weights, n_jobs=3, remove_duplicates=True)

    assert np.allclose(reference_activations, activations)
    assert np.allclose(reference_activations, activations_mp)


def test_ignore_missing_cues():
    weights = xr.DataArray(np.array([[0, 1, 0], [1, 0, 0]]),
                           coords={
                               'outcomes': ['o1', 'o2'],
                               'cues': ['c1', 'c2', 'c3']
                           },
                           dims=('outcomes', 'cues'))

    events = [(['c1', 'c2', 'c3'], []),
              (['c1', 'c3'], []),
              (['c2', 'c4'], []),
              (['c1', 'c1'], [])]
    reference_activations = np.array([[1, 0, 1, 0], [1, 1, 0, 1]])

    with pytest.raises(KeyError):
        activations = activation(events, weights, n_jobs=1,
                                 remove_duplicates=True)

    activations = activation(events, weights, n_jobs=1,
                             remove_duplicates=True, ignore_missing_cues=True)
    activations_mp = activation(events, weights, n_jobs=3,
                                remove_duplicates=True, ignore_missing_cues=True)

    assert np.allclose(reference_activations, activations)
    assert np.allclose(reference_activations, activations_mp)


def test_activation_dict():
    weights = defaultdict(lambda: defaultdict(float))
    weights['o1']['c1'] = 0
    weights['o1']['c2'] = 1
    weights['o1']['c3'] = 0
    weights['o2']['c1'] = 1
    weights['o2']['c2'] = 0
    weights['o2']['c3'] = 0
    events = [(['c1', 'c2', 'c3'], []),
              (['c1', 'c3'], []),
              (['c2'], []),
              (['c1', 'c1'], [])]
    reference_activations = {
        'o1': [1, 0, 1, 0],
        'o2': [1, 1, 0, 1]
    }

    with pytest.raises(ValueError):
        activations = activation(events, weights, n_jobs=1)

    activations = activation(events, weights, n_jobs=1, remove_duplicates=True)
    for outcome, activation_list in activations.items():
        assert np.allclose(reference_activations[outcome], activation_list)


def test_ignore_missing_cues_dict():
    weights = defaultdict(lambda: defaultdict(float))
    weights['o1']['c1'] = 0
    weights['o1']['c2'] = 1
    weights['o1']['c3'] = 0
    weights['o2']['c1'] = 1
    weights['o2']['c2'] = 0
    weights['o2']['c3'] = 0
    events = [(['c1', 'c2', 'c3'], []),
              (['c1', 'c3'], []),
              (['c2', 'c4'], []),
              (['c1', 'c1'], [])]
    reference_activations = {
        'o1': [1, 0, 1, 0],
        'o2': [1, 1, 0, 1]
    }

    with pytest.raises(ValueError):
        activations = activation(events, weights, n_jobs=1)

    activations = activation(events, weights, n_jobs=1,
                             remove_duplicates=True, ignore_missing_cues=True)
    for outcome, activation_list in activations.items():
        assert np.allclose(reference_activations[outcome], activation_list)


@pytest.mark.runslow
def test_activation_matrix_large():
    """
    Test with a lot of data. Better run only with at least 12GB free RAM.
    To get time prints for single and multiprocessing run with pytest ... --capture=no --runslow
    """
    print("")
    print("Start setup...")

    def time_test(func, of=""):  # pylint: disable=invalid-name
        def dec_func(*args, **kwargs):
            print("start test '{}'".format(of))
            start = time.clock()
            res = func(*args, **kwargs)
            end = time.clock()
            print("finished test '{}'".format(of))
            print("  duration: {:.3f}s".format(end - start))
            print("")
            return res
        return dec_func

    nn = 2000
    n_cues = 10*nn
    n_outcomes = nn
    n_events = 10*nn
    n_cues_per_event = 30
    weight_mat = np.random.rand(n_cues, n_outcomes)
    cues = ['c'+str(i) for i in range(n_cues)]
    weights = xr.DataArray(weight_mat,
                           coords={'cues': cues},
                           dims=('cues', 'outcomes'))
    events = [(np.random.choice(cues, n_cues_per_event), [])
              for i in range(n_events)]  # no generator, we use it twice

    print("Start test...")
    print("")
    gc.collect()
    asp = (time_test(activation, of="single threaded")
           (events, weights, n_jobs=1, remove_duplicates=True))
    gc.collect()
    amp = (time_test(activation, of="multi threaded (up to 8 threads)")
           (events, weights, n_jobs=8, remove_duplicates=True))
    del weights
    del events
    gc.collect()
    print("Compare results...")
    assert np.allclose(asp, amp), "single and multi threaded had different results"
    print("Equal.")



"""
pyndl.preprocess
----------------

*pyndl.preprocess* provides functions in order to preprocess data and create
event files from it.

"""
import collections
import gzip
import multiprocessing
import os
import random
import re
import sys
import time


def bandsample(population, sample_size=50000, *, cutoff=5, seed=None,
               verbose=False):
    """
    Creates a sample of size sample_size out of the population using
    band sampling.

    """
    # make a copy of the population
    # filter all words with freq < cutoff
    population = [(word, freq) for word, freq in population.items() if freq >=
                  cutoff]

    if seed is not None:
        raise NotImplementedError("Reproducable bandsamples by seeding are not properly implemented yet.")

    # shuffle words with same frequency
    rand = random.Random(seed)
    rand.shuffle(population)
    population.sort(key=lambda x: x[1])  # lowest -> highest freq

    step = sum(freq for word, freq in population) / sample_size
    if verbose:
        print("step %.2f" % step)

    accumulator = 0
    index = 0
    sample = list()
    while 0 <= index < len(population):
        word, freq = population[index]
        accumulator += freq
        if verbose:
            print("%s\t%i\t%.2f" % (word, freq, accumulator))
        if accumulator >= step:
            sample.append((word, freq))
            accumulator -= step
            if verbose:
                print("add\t%s\t%.2f" % (word, accumulator))
            del population[index]
            while accumulator >= step and index >= 1:
                index -= 1
                sample.append(population[index])
                accumulator -= step
                if verbose:
                    word, freq = population[index]
                    print("  add\t%s\t%.2f" % (word, accumulator))
                del population[index]
        else:
            # only add to index if no element was removed
            # if element was removed, index points at next element already
            index += 1
            if verbose and index % 1000 == 0:
                print(".", end="")
                sys.stdout.flush()
    sample = collections.Counter({key: value for key, value in sample})
    return sample


def ngrams_to_word(occurrences, n_chars, outfile, remove_duplicates=True):
    """
    Process the occurrences and write them to outfile.

    Parameters
    ----------
    occurrences : sequence of (cues, outcomes) tuples
        cues and outcomes are both strings where underscores and # are
        special symbols.
    n_chars : number of characters (e.g. 2 for bigrams, 3 for trigrams, ...)
    outfile : file handle

    remove_duplicates : bool
        if True make cues and outcomes per event unique

    """
    for cues, outcomes in occurrences:
        if cues and outcomes:
            occurrence = cues + '_' + outcomes
        else:  # take either
            occurrence = cues + outcomes
        phrase_string = "#" + re.sub("_", "#", occurrence) + "#"
        ngrams = (phrase_string[i:(i + n_chars)] for i in
                  range(len(phrase_string) - n_chars + 1))
        if not ngrams or not occurrence:
            continue
        if remove_duplicates:
            ngrams = set(ngrams)
            occurrence = "_".join(set(occurrence.split("_")))
        outfile.write("{}\t{}\n".format("_".join(ngrams), occurrence))


def process_occurrences(occurrences, outfile, *,
                        cue_structure="trigrams_to_word", remove_duplicates=True):
    """
    Process the occurrences and write them to outfile.

    Parameters
    ----------
    occurrences : sequence of (cues, outcomes) tuples
        cues and outcomes are both strings where underscores and # are
        special symbols.
    outfile : file handle

    cue_structure : {'bigrams_to_word', 'trigrams_to_word', 'word_to_word'}

    remove_duplicates : bool
        if True make cues and outcomes per event unique

    """
    if cue_structure == "bigrams_to_word":
        ngrams_to_word(occurrences, 2, outfile, remove_duplicates=remove_duplicates)
    elif cue_structure == "trigrams_to_word":
        ngrams_to_word(occurrences, 3, outfile, remove_duplicates=remove_duplicates)
    elif cue_structure == "word_to_word":
        for cues, outcomes in occurrences:
            if not cues:
                continue
            if remove_duplicates:
                cues = "_".join(set(cues.split("_")))
                outcomes = "_".join(set(outcomes.split("_")))
            outfile.write("{}\t{}\n".format(cues, outcomes))
    else:
        raise NotImplementedError('cue_structure=%s is not implemented yet.' % cue_structure)


def create_event_file(corpus_file,
                      event_file,
                      symbols="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
                      *,
                      context_structure="document",
                      event_structure="consecutive_words",
                      event_options=(3,),  # number_of_words,
                      cue_structure="trigrams_to_word",
                      lower_case=False,
                      remove_duplicates=True,
                      verbose=False):
    """
    Create an text based event file from a corpus file.

    Parameters
    ----------
    corpus_file : str
        path where the corpus file is
    event_file : str
        path where the output file will be created
    symbols : str
        string of all valid symbols
    context_structure : {"document", "paragraph", "line"}

    event_structure : {"line", "consecutive_words", "word_to_word", "sentence"}

    event_options : None or (number_of_words,) or (before, after) or None
        in "consecutive words" the number of words of the sliding window as
        an integer; in "word_to_word" the number of words before and after the
        word of interest each as an integer.
    cue_structure: {"trigrams_to_word", "word_to_word", "bigrams_to_word"}

    lower_case : bool
        should the cues and outcomes be lower cased
    remove_duplicates : bool
        create unique cues and outcomes per event
    verbose : bool

    Notes
    -----
    Breaks / Separators :

        What marks parts, where we do not want to continue learning?

        * ``---end.of.document---`` string?
        * line breaks?
        * empty lines?

        What do we consider one event?

        * three consecutive words?
        * one line of the corpus?
        * everything between two empty lines?
        * everything within one document?

        Should the events be connected to the events before and after?

        No.

    Context :

        A context is a whole document or a paragraph within which we will take
        (three) consecutive words as occurrences or events. The last words of a
        context will not form an occurrence with the first words of the next
        context.

    Occurrence :

        An occurrence or event is will result in one event in the end. This can
        be (three) consecutive words, a sentence, or a line in the corpus file.

    """
    if '_' in symbols or '#' in symbols or '\t' in symbols:
        raise ValueError('"_", "#", and "\\t" are special symbols and cannot be in symbols string')

    if event_structure not in ('consecutive_words', 'line', 'word_to_word'):
        raise NotImplementedError('This event structure (%s) is not implemented yet.' % event_structure)

    if context_structure not in ('document', 'line'):
        raise NotImplementedError('This context structure (%s) is not implemented yet.' % context_structure)

    if os.path.isfile(event_file):
        raise OSError('%s file exits. Remove file and start again.' % event_file)

    # in_symbols = re.compile("^[%s]*$" % symbols)
    not_in_symbols = re.compile("[^%s]" % symbols)
    context_pattern = re.compile("(---end.of.document---|---END.OF.DOCUMENT---)")

    if event_structure == 'consecutive_words':
        number_of_words, = event_options
    elif event_structure == 'word_to_word':
        before, after = event_options

    def gen_occurrences(words):
        """
        Make an occurrence out of consecutive words.

        Take all number_of_words number of consecutive words and make an
        occurrence out of it.

        For words = (A, B, C, D); number_of_words = 3 make: (A, ), (A_B, ),
        (A_B_C, ), (B_C_D, ), (C_D, ), (D, )

        """
        if event_structure == 'consecutive_words':
            occurrences = list()
            # can't have more consecutive words than total words
            length = min(number_of_words, len(words))
            # slide window over list of words
            for ii in range(1 - length, len(words)):
                # no consecutive words before first word
                start = max(ii, 0)
                # no consecutive words after last word
                end = min(ii + length, len(words))
                # append (cues, outcomes) with empty outcomes
                occurrences.append(("_".join(words[start:end]), ""))
            return occurrences
        # for words = (A, B, C, D); before = 2, after = 1
        # make: (B, A), (A_C, B), (A_B_D, C), (B_C, D)
        elif event_structure == 'word_to_word':
            occurrences = list()
            for ii, word in enumerate(words):
                # words before the word to a maximum of before
                cues = words[max(0, ii - before):ii]
                # words after the word to a maximum of before
                cues.extend(words[(ii + 1):min(len(words), ii + 1 + after)])
                # append (cues, outcomes)
                occurrences.append(("_".join(cues), word))
            return occurrences
        elif event_structure == 'line':
            # (cues, outcomes) with empty outcomes
            return [('_'.join(words), ''), ]
        else:
            raise ValueError('gen_occurrences should be one of {"consecutive_words", "word_to_word", "line"}')

    def process_line(line):
        """processes one line of text."""
        if lower_case:
            line = line.lower()
        # replace all weird characters with space
        line = not_in_symbols.sub(" ", line)
        return line

    def gen_words(line):
        """generates words out of a line of text."""
        return [word.strip() for word in line.split(" ") if word.strip()]

    def process_words(words):
        """processes one word and makes an occurrence out of it."""
        occurrences = gen_occurrences(words)
        process_occurrences(occurrences, outfile,
                            cue_structure=cue_structure,
                            remove_duplicates=remove_duplicates)

    def process_context(line):
        """called when a context boundary is found."""
        if context_structure == 'document':
            # remove document marker
            line = context_pattern.sub("", line)
        return line

    with open(corpus_file, "rt") as corpus:
        with gzip.open(event_file, "wt") as outfile:
            outfile.write("cues\toutcomes\n")

            words = []
            for ii, line in enumerate(corpus):
                if verbose and ii % 100000 == 0:
                    print(".", end="")
                    sys.stdout.flush()
                    outfile.flush()
                line = line.strip()

                if context_structure == 'line':
                    line = process_line(line)
                    words = gen_words(line)
                    process_words(words)
                else:
                    if context_pattern.search(line) is not None:
                        # process the first context
                        context1, *contexts = context_pattern.split(line)
                        context1 = process_context(context1)

                        if context1.strip():
                            context1 = process_line(context1.strip())
                            words.extend(gen_words(context1))
                        process_words(words)
                        # process in between contexts
                        while len(contexts) > 1:
                            words = []
                            context1, *contexts = contexts
                            context1 = process_context(context1)
                            if context1.strip():
                                context1 = process_line(context1.strip())
                                words.extend(gen_words(context1))
                                process_words(words)
                        # add last part to next context
                        context1 = contexts[0]
                        context1 = process_context(context1)
                        if context1.strip():
                            context1 = process_line(context1.strip())
                            words.extend(gen_words(context1))
                    else:
                        line = process_line(line)
                        words.extend(gen_words(line))

            # write the last context (the rest) when context_structure is not
            # 'line'
            if context_structure != 'line':
                process_words(words)


class JobFilter():
    # pylint: disable=E0202,missing-docstring

    """
    Stores the persistent information over several jobs and exposes a job
    method that only takes the varying parts as one argument.

    .. note::

        Using a closure is not possible as it is not pickable / serializable.

    """

    @staticmethod
    def return_empty_string():
        return ''

    def __init__(self, keep_cues, keep_outcomes, remove_cues, remove_outcomes,
                 cue_map, outcome_map):
        if ((cue_map is not None and remove_cues is not None) or
                (cue_map is not None and keep_cues != 'all') or
                (remove_cues is not None and keep_cues != 'all')):
            raise ValueError('You can either specify keep_cues, remove_cues, or cue_map.')
        if ((outcome_map is not None and remove_outcomes is not None) or
                (outcome_map is not None and keep_outcomes != 'all') or
                (remove_outcomes is not None and keep_outcomes != 'all')):
            raise ValueError('You can either specify keep_outcomes, remove_outcomes, or outcome_map.')

        if cue_map is not None:
            self.cue_map = collections.defaultdict(self.return_empty_string, cue_map)
            self.process_cues = self.process_cues_map
        elif remove_cues is not None:
            self.remove_cues = set(remove_cues)
            self.process_cues = self.process_cues_remove
        elif keep_cues == 'all':
            self.keep_cues = 'all'
            self.process_cues = self.process_cues_all
        else:
            self.keep_cues = keep_cues
            self.process_cues = self.process_cues_keep
        if outcome_map is not None:
            self.outcome_map = collections.defaultdict(self.return_empty_string, outcome_map)
            self.process_outcomes = self.process_outcomes_map
        elif remove_outcomes is not None:
            self.remove_outcomes = set(remove_outcomes)
            self.process_outcomes = self.process_outcomes_remove
        elif keep_outcomes == 'all':
            self.keep_outcomes = 'all'
            self.process_outcomes = self.process_outcomes_all
        else:
            self.keep_outcomes = set(keep_outcomes)
            self.process_outcomes = self.process_outcomes_keep

    def process_cues(self, cues):
        raise NotImplementedError("Needs to be implemented or assigned by a specific method.")

    def process_cues_map(self, cues):
        cues = [self.cue_map[cue] for cue in cues]
        return [cue for cue in cues if cue]

    def process_cues_remove(self, cues):
        return [cue for cue in cues if cue not in self.remove_cues]

    def process_cues_keep(self, cues):
        return [cue for cue in cues if cue in self.keep_cues]

    def process_cues_all(self, cues):
        return cues

    def process_outcomes(self, outcomes):
        raise NotImplementedError("Needs to be implemented or assigned by a specific method.")

    def process_outcomes_map(self, outcomes):
        outcomes = [self.outcome_map[outcome] for outcome in outcomes]
        return [outcome for outcome in outcomes if outcome]

    def process_outcomes_remove(self, outcomes):
        return [outcome for outcome in outcomes if outcome not in self.remove_outcomes]

    def process_outcomes_keep(self, outcomes):
        return [outcome for outcome in outcomes if outcome in self.keep_outcomes]

    def process_outcomes_all(self, outcomes):
        return outcomes

    def job(self, line):
        try:
            cues, outcomes = line.strip('\n').split("\t")
        except ValueError:
            raise ValueError("tabular event file need to have two tab separated columns")
        cues = cues.split("_")
        outcomes = outcomes.split("_")
        cues = self.process_cues(cues)
        outcomes = self.process_outcomes(outcomes)
        # no cues left?
        # NOTE: We want to keep events with no outcomes as this is the
        # background for the cues in that events.
        if not cues:
            return None
        processed_line = ("%s\t%s\n" % ("_".join(cues), "_".join(outcomes)))
        return processed_line


def filter_event_file(input_event_file, output_event_file, *,
                      keep_cues="all", keep_outcomes="all",
                      remove_cues=None, remove_outcomes=None,
                      cue_map=None, outcome_map=None,
                      n_jobs=1, number_of_processes=None, chunksize=100000,
                      verbose=False):
    """
    Filter an event file by a list or a map of cues and outcomes.

    Parameters
    ----------
    You can either use keep_*, remove_*, or map_*.

    input_event_file : str
        path where the input event file is
    output_event_file : str
        path where the output file will be created
    keep_cues : "all" or sequence of str
        list of all cues that should be kept
    keep_outcomes : "all" or sequence of str
        list of all outcomes that should be kept
    remove_cues : None or sequence of str
        list of all cues that should be removed
    remove_outcomes : None or sequence of str
        list of all outcomes that should be removed
    cue_map : dict
        maps every cue as key to the value. Removes all cues that do not have a
        key. This can be used to map several different cues to the same cue or
        to rename cues.
    outcome_map : dict
        maps every outcome as key to the value. Removes all outcome that do not have a
        key. This can be used to map several different outcomes to the same
        outcome or to rename outcomes.
    n_jobs : int
        number of threads to use
    chunksize : int
        number of chunks per submitted job, should be around 100000

    Notes
    -----
    It will keep all cues that are within the event and that (for a human
    reader) might clearly belong to a removed outcome. This is on purpose and
    is the expected behaviour as these cues are in the context of this outcome.

    If an event has no cues it gets removed, but if an event has no outcomes it
    is still present in order to capture the background rate of that cues.

    """
    if number_of_processes is not None:
        warnings.warn("Parameter `number_of_processes` is renamed to `n_jobs`. The old name "
                      "will stop working with v0.9.0.",
                      DeprecationWarning, stacklevel=2)
        n_jobs = number_of_processes
    job = JobFilter(keep_cues, keep_outcomes, remove_cues, remove_outcomes,
                    cue_map, outcome_map)

    with multiprocessing.Pool(n_jobs) as pool:
        with gzip.open(input_event_file, "rt") as infile:
            with gzip.open(output_event_file, "wt") as outfile:
                # copy header
                outfile.write(infile.readline())
                for ii, processed_line, in enumerate(pool.imap(job.job, infile,
                                                               chunksize=chunksize)):
                    if processed_line is not None:
                        outfile.write(processed_line)
                    if verbose and ii % 100000 == 0:
                        print('.', end='')
                        sys.stdout.flush()


################
#  Preprocessing
################

MAGIC_NUMBER = 14159265
CURRENT_VERSION_WITH_FREQ = 215
CURRENT_VERSION = 2048 + 215


def read_binary_file(binary_file_path):
    with open(binary_file_path, "rb") as binary_file:
        magic_number = to_integer(binary_file.read(4))
        if not magic_number == MAGIC_NUMBER:
            raise ValueError('Header does not match the magic number')
        version = to_integer(binary_file.read(4))
        if version == CURRENT_VERSION:
            pass
        else:
            raise ValueError('Version is incorrectly specified')

        nr_of_events = to_integer(binary_file.read(4))
        for _ in range(nr_of_events):
            # Cues
            number_of_cues = to_integer(binary_file.read(4))
            cue_ids = [to_integer(binary_file.read(4)) for ii in range(number_of_cues)]
            # outcomes
            number_of_outcomes = to_integer(binary_file.read(4))
            outcome_ids = [to_integer(binary_file.read(4)) for ii in range(number_of_outcomes)]
            yield (cue_ids, outcome_ids)


def to_bytes(int_):
    return int_.to_bytes(4, 'little')


def to_integer(byte_):
    return int.from_bytes(byte_, "little")


def write_events(events, filename, *, start=0, stop=4294967295, remove_duplicates=None):
    """
    Write out a list of events to a disk file in binary format.

    Parameters
    ----------
    events : iterator of (cue_ids, outcome_ids) tuples called event

    filename : string

    start : first event to write (zero based index)

    stop : last event to write (zero based index; excluded)

    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)

    Returns
    -------
    number_events : int
        actual number of events written to file

    Notes
    -----
    The **binary format** as the following structure::

        8 byte header
        nr of events
        nr of cues in first event
        ids for every cue
        nr of outcomes in first event
        ids for every outcome
        nr of cues in second event
        ...

    Raises
    ------
    StopIteration : events generator is exhausted before stop is reached

    """
    with open(filename, "wb") as out_file:
        # 8 bytes header
        out_file.write(to_bytes(MAGIC_NUMBER))
        out_file.write(to_bytes(CURRENT_VERSION))

        # events
        # estimated number of events (will be rewritten if the actual number
        # differs)
        n_events_estimate = stop - start
        out_file.write(to_bytes(n_events_estimate))

        n_events = 0

        for ii, event in enumerate(events):
            if ii < start:
                continue
            if ii >= stop:
                break

            n_events += 1
            cue_ids, outcome_ids = event

            if remove_duplicates is None:
                if len(cue_ids) != len(set(cue_ids)) or len(outcome_ids) != len(set(outcome_ids)):
                    raise ValueError(''.join([
                        'event %i does not have unique cues or outcomes.'
                        'Use remove_duplicates=True in order to force unique cues and outcomes.'
                        'Use remove_duplicates=False to allow the same cue or outcome multiple'
                        'times in the same event (not recommended)']) % ii)
            elif remove_duplicates:
                cue_ids = set(cue_ids)
                outcome_ids = set(outcome_ids)
            else:
                pass

            # cues in event
            out_file.write(to_bytes(len(cue_ids)))
            for cue_id in cue_ids:
                out_file.write(to_bytes(cue_id))

            # outcomes in event
            out_file.write(to_bytes(len(outcome_ids)))
            for outcome_id in outcome_ids:
                out_file.write(to_bytes(outcome_id))

        if n_events != n_events_estimate and not n_events == 0:
            # the generator was exhausted earlier
            out_file.seek(8)
            out_file.write(to_bytes(n_events))
            raise StopIteration(("event generator was exhausted before stop", n_events))

    if n_events == 0:
        os.remove(filename)
    return n_events


def event_generator(event_file, cue_id_map, outcome_id_map, *, sort_within_event=False):
    with gzip.open(event_file, "rt") as in_file:
        # skip header
        in_file.readline()
        for _, line in enumerate(in_file):
            try:
                cues, outcomes = line.strip('\n').split("\t")
            except ValueError:
                raise ValueError("tabular event file need to have three tab separated columns")
            cues = cues.split("_")
            outcomes = outcomes.split("_")
            # uses list and not generators; as generators can only be traversed once
            event = ([cue_id_map[cue] for cue in cues],
                     [outcome_id_map[outcome] for outcome in outcomes])
            if sort_within_event:
                cue_ids, outcome_ids = event
                cue_ids = list(cue_ids)
                cue_ids.sort()
                outcome_ids = list(outcome_ids)
                outcome_ids.sort()
                event = (cue_ids, outcome_ids)
            yield event


def _job_binary_event_file(*,
                           file_name,
                           event_file,
                           cue_id_map,
                           outcome_id_map,
                           sort_within_event,
                           start,
                           stop,
                           remove_duplicates):
    # create generator which is not pickable
    events = event_generator(event_file, cue_id_map, outcome_id_map, sort_within_event=sort_within_event)
    n_events = write_events(events, file_name, start=start, stop=stop, remove_duplicates=remove_duplicates)
    return n_events


def create_binary_event_files(event_file,
                              path_name,
                              cue_id_map,
                              outcome_id_map,
                              *,
                              sort_within_event=False,
                              n_jobs=2,
                              events_per_file=10000000,
                              overwrite=False,
                              remove_duplicates=None,
                              verbose=False):
    """
    Creates the binary event files for a tabular cue outcome corpus.

    Parameters
    ----------
    event_file : str
        path to tab separated text file that contains all events in a cue
        outcome table.
    path_name : str
        folder name where to store the binary event files
    cue_id_map : dict (str -> int)
        cue to id map
    outcome_id_map : dict (str -> int)
        outcome to id map
    sort_within_event : bool
        should we sort the cues and outcomes within the event
    n_jobs : int
        number of threads to use
    events_per_file : int
        Number of events in each binary file. Has to be larger than 1
    overwrite : bool
        overwrite files if they exist
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    verbose : bool

    Returns
    -------
    number_events : int
        sum of number of events written to binary files
    """
    # pylint: disable=missing-docstring

    if events_per_file < 2:
        raise ValueError("events_per_file has to be larger than 1")

    if not os.path.isdir(path_name):
        if verbose:
            print("create event file folder '%s'" % path_name)
        os.mkdir(path_name, 0o773)
    elif not overwrite:
        raise IOError("folder %s exists and overwrite is False" % path_name)
    else:
        if verbose:
            print("removing event files in '%s'" % path_name)
        for file_name in os.listdir(path_name):
            if "events_0_" in file_name:
                os.remove(os.path.join(path_name, file_name))

    number_events = 0

    with multiprocessing.Pool(n_jobs) as pool:

        def _error_callback(error):
            if isinstance(error, StopIteration):
                _, result = error.value
                nonlocal number_events
                number_events += result  # pylint: disable=undefined-variable
                pool.close()
            else:
                raise error

        def _callback(result):
            nonlocal number_events
            number_events += result
            if verbose:
                print("finished job")
                sys.stdout.flush()

        ii = 0
        while True:
            kwargs = {
                "file_name": os.path.join(path_name, "events_0_%i.dat" % ii),
                "event_file": event_file,
                "cue_id_map": cue_id_map,
                "outcome_id_map": outcome_id_map,
                "sort_within_event": sort_within_event,
                "start": ii*events_per_file,
                "stop": (ii+1)*events_per_file,
                "remove_duplicates": remove_duplicates,
            }
            try:
                result = pool.apply_async(_job_binary_event_file,
                                          kwds=kwargs,
                                          callback=_callback,
                                          error_callback=_error_callback)
                if verbose:
                    print("submitted job %i" % ii)
            except ValueError as error:
                # someone has closed the pool with the correct error callback
                if error.args[0] == 'Pool not running':
                    if verbose:
                        print("reached end of events")
                    break  # out of while True
                else:
                    raise error
            ii += 1
            # only start jobs in chunks of 4*n_jobs
            if ii % (n_jobs*4) == 0:
                while True:
                    if result.ready():
                        break
                    time.sleep(1.0)  # check every second
                    if verbose:
                        print('c')
                        sys.stdout.flush()
        # wait until all jobs are done
        pool.close()
        pool.join()
        if verbose:
            print("finished all jobs.\n")
    return number_events

# for example code see function test_preprocess in file
# ./tests/test_preprocess.py.



from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext
import sys


# bootstrap numpy
# https://stackoverflow.com/questions/19919905/how-to-bootstrap-numpy-installation-in-setup-py
class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())


pkg = __import__('pyndl')

author =  pkg.__author__
email = pkg.__author_email__

version = pkg.__version__
classifiers = pkg.__classifiers__

description = pkg.__description__

def load_requirements(fn):
    """Read a requirements file and create a list that can be used in setup."""
    with open(fn, 'r') as f:
        return [x.rstrip() for x in list(f) if x and not x.startswith('#')]


ndl_parallel = Extension("pyndl.ndl_parallel", ["pyndl/ndl_parallel.pyx"])
ndl_openmp = Extension("pyndl.ndl_openmp", ["pyndl/ndl_openmp.pyx"],
                       extra_compile_args=['-fopenmp'], extra_link_args=['-fopenmp'])
corr_parallel = Extension("pyndl.correlation_openmp", ["pyndl/correlation_openmp.pyx"],
                       extra_compile_args=['-fopenmp'], extra_link_args=['-fopenmp'])
# by giving ``cython`` as ``install_requires`` this will be ``cythonized``
# automagically

ext_modules = []
if sys.platform.startswith('linux'):
    ext_modules = [ndl_parallel, ndl_openmp, corr_parallel]
elif sys.platform.startswith('win32'):
    ext_modules = [ndl_parallel] # skip openmp installation on windows for now
elif sys.platform.startswith('darwin'):
    ext_modules = [ndl_parallel]  # skip openmp installation on macos for now

setup(
    name='pyndl',
    version=version,
    license='MIT',
    description=description,
    long_description=open('README.rst', encoding="utf-8").read(),
    author=author,
    author_email=email,
    url='https://github.com/quantling/pyndl',
    classifiers=classifiers,
    platforms='Linux',
    packages=['pyndl'],
    setup_requires=['numpy', 'cython'],
    install_requires=load_requirements('requirements.txt'),
    extras_require={
        'tests': [
            'pylint',
            'pytest',
            'pycodestyle'],
        'docs': [
            'sphinx >= 1.4',
            'sphinx_rtd_theme',
            'numpydoc',
            'easydev==0.9.35']},
    ext_modules=ext_modules,
    cmdclass={'build_ext': build_ext}
)



cimport numpy as np
ctypedef np.float64_t dtype_t
from error_codes cimport ErrorCode


cdef ErrorCode learn_inplace_binary_to_binary_ptr(char*, dtype_t, dtype_t, dtype_t, dtype_t,
                        dtype_t*, unsigned int, unsigned int*, unsigned int,
                        unsigned int) nogil


cdef ErrorCode learn_inplace_binary_to_real_ptr(char*, dtype_t, dtype_t*,
                        dtype_t*, unsigned int, unsigned int, unsigned int,
                        unsigned int) nogil


cdef ErrorCode learn_inplace_real_to_real_ptr(char*, dtype_t, dtype_t*,
                        dtype_t*, dtype_t*, unsigned int, unsigned int,
                        unsigned int, unsigned int) nogil


cdef ErrorCode learn_inplace_real_to_binary_ptr(char*, dtype_t, dtype_t,
                        dtype_t, dtype_t*, dtype_t*, unsigned int, unsigned
                        int, unsigned int) nogil



"""
pyndl.count
-----------

*pyndl.count* provides functions in order to count

* words and symbols in a corpus file
* cues and outcomes in an event file

"""
# pylint: disable=redefined-outer-name, invalid-name

from collections import Counter, namedtuple
import gzip
import itertools
import multiprocessing
import sys
import warnings


CuesOutcomes = namedtuple('CuesOutcomes', 'n_events, cues, outcomes')
WordsSymbols = namedtuple('WordsSymbols', 'words, symbols')


def _job_cues_outcomes(event_file_name, start, step, verbose=False):
    """
    Counts cues and outcomes for every ``step`` event starting from
    ``start`` event.

    Returns
    -------
    (nn, cues, outcomes) : (int, collections.Counter, collections.Counter)

    """
    cues = Counter()
    outcomes = Counter()
    nn = -1  # in case the for loop never gets called and 1 gets added in the end
    with gzip.open(event_file_name, 'rt') as dfile:
        # skip header
        dfile.readline()
        for nn, line in enumerate(itertools.islice(dfile, start, None, step)):
            cues_line, outcomes_line = line.split('\t')
            for cue in cues_line.split('_'):
                cues[cue] += 1
            for outcome in outcomes_line.strip().split('_'):
                outcomes[outcome] += 1
            if verbose and nn % 100000 == 0:
                print('.', end='')
                sys.stdout.flush()
    return (nn + 1, cues, outcomes)


def cues_outcomes(event_file_name,
                  *, n_jobs=2, number_of_processes=None, verbose=False):
    """
    Counts cues and outcomes in event_file_name using n_jobs
    processes.

    Returns
    -------
    (n_events, cues, outcomes) : (int, collections.Counter, collections.Counter)

    """
    if number_of_processes is not None:
        warnings.warn("Parameter `number_of_processes` is renamed to `n_jobs`. The old name "
                      "will stop working with v0.9.0.",
                      DeprecationWarning, stacklevel=2)
        n_jobs = number_of_processes
    with multiprocessing.Pool(n_jobs) as pool:
        step = n_jobs
        results = pool.starmap(_job_cues_outcomes,
                               ((event_file_name,
                                 start,
                                 step,
                                 verbose)
                                for start in range(n_jobs)))
        n_events = 0
        cues = Counter()
        outcomes = Counter()
        for nn, cues_process, outcomes_process in results:
            n_events += nn
            cues += cues_process
            outcomes += outcomes_process

    if verbose:
        print('\n...counting done.')

    return CuesOutcomes(n_events, cues, outcomes)


def _job_words_symbols(corpus_file_name, start, step, lower_case=False,
                       verbose=False):
    """
    Counts the words and symbols for every ``step`` line starting from
    ``start`` line.

    It is assumed that words are separated by at least one space or by a new
    line character.

    .. note::

        Punctuation characters, brackets and some other characters are stripped
        from the word and are not counted.

    Returns
    -------
    (words, symbols) : (collections.Counter, collections.Counter)

    """
    words = Counter()
    symbols = Counter()
    with open(corpus_file_name, 'r') as dfile:
        for nn, line in enumerate(itertools.islice(dfile, start, None, step)):
            for word in line.split():  # splits the string on all whitespace
                word = word.strip()
                word = word.strip('!?,.:;/"\'()^@*~')
                if lower_case:
                    word = word.lower()
                if not word:
                    continue
                words[word] += 1
                symbols += Counter(word)
            if verbose and nn % 100000 == 0:
                print('.', end='')
                sys.stdout.flush()
    return (words, symbols)


def words_symbols(corpus_file_name,
                  *, n_jobs=2, number_of_processes=None, lower_case=False, verbose=False):
    """
    Counts words and symbols in corpus_file_name using n_jobs
    processes.

    Returns
    -------
    (words, symbols) : (collections.Counter, collections.Counter)

    """
    if number_of_processes is not None:
        warnings.warn("Parameter `number_of_processes` is renamed to `n_jobs`. The old name "
                      "will stop working with v0.9.0.",
                      DeprecationWarning, stacklevel=2)
        n_jobs = number_of_processes
    with multiprocessing.Pool(n_jobs) as pool:
        step = n_jobs
        results = pool.starmap(_job_words_symbols, ((corpus_file_name,
                                                     start,
                                                     step,
                                                     lower_case,
                                                     verbose)
                                                    for start in
                                                    range(n_jobs)))
        words = Counter()
        symbols = Counter()
        for words_process, symbols_process in results:
            words += words_process
            symbols += symbols_process

    if verbose:
        print('\n...counting done.')

    return WordsSymbols(words, symbols)


def save_counter(counter, filename, *, header='key\tfreq\n'):
    """
    Saves a counter object into a tab delimitered text file.

    """
    with open(filename, 'wt') as dfile:
        dfile.write(header)
        for key, count in counter.most_common():
            dfile.write('{key}\t{count}\n'.format(key=key, count=count))


def load_counter(filename):
    """
    Loads a counter out of a tab delimitered text file.

    """
    with open(filename, 'rt') as dfile:
        # skip header
        dfile.readline()
        counter = Counter()
        for line in dfile:
            key, count = line.strip().split('\t')
            if key in counter.keys():
                raise ValueError("%s contains two instances (words, symbols, ...) of the same spelling." % filename)
            counter[key] = int(count)
    return counter



"""
pyndl.wh
--------

*pyndl.wh* provides functions in order to train WH models

"""
from collections import defaultdict, OrderedDict
import copy
import getpass
import os
# from queue import Queue
import socket
import sys
import tempfile
# import threading
import time

import cython
import pandas as pd
import numpy as np
import xarray as xr

from . import __version__ as pyndl_version
from . import count
from . import preprocess
from . import io
from . import ndl

# conditional import as openmp is only compiled for linux
if sys.platform.startswith('linux'):
    from . import ndl_openmp
elif sys.platform.startswith('win32'):
    pass
elif sys.platform.startswith('darwin'):
    pass


WeightDict = ndl.WeightDict


def wh(events, eta, *, cue_vectors=None, outcome_vectors=None,
        method='openmp', weights=None,
        n_jobs=8, n_outcomes_per_job=10, remove_duplicates=None,
        verbose=False, temporary_directory=None,
        events_per_temporary_file=10000000):
    """
    Calculate the weights for all events using the Widrow-Hoff learning rule
    and training as outcomes on sematic vectors in semantics.

    This is a parallel python implementation using numpy, multithreading and
    the binary format defined in preprocess.py.

    Parameters
    ----------
    events : str
        path to the event file
    eta : float
        learning rate
    cue_vectors : xarray.DataArray
        matrix that contains the cue vectors for each cue
    outcome_vectors : xarray.DataArray
        matrix that contains the target vectors for each outcome
    method : {'openmp', 'threading', 'numpy'}
        'numpy' works only for real to real Widrow-Hoff.
    weights : None or xarray.DataArray
        the xarray.DataArray needs to have the dimensions 'cues' and 'outcomes'
    n_jobs : int
        an integer giving the number of threads in which the job should be
        executed
    n_outcomes_per_job : int
        an integer giving the number of outcomes that are processed in one job
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    verbose : bool
        print some output if True.
    temporary_directory : str
        path to directory to use for storing temporary files created;
        if none is provided, the operating system's default will
        be used (/tmp on unix)
    events_per_temporary_file: int
        Number of events in each temporary binary file. Has to be larger than 1

    Returns
    -------
    weights : xarray.DataArray
        the dimensions of the weights reflect the type of Widrow-Hoff that was
        run (real to real, binary to real, real to binary or binary to binary).
        The dimension names reflect this in the weights. They are a combination
        of 'outcomes' x 'outcome_vector_dimensions' and 'cues' x
        'cue_vector_dimensions'
        with dimensions 'outcome_vector dimensions' and
        'cue_vector_dimensions'. You can lookup the weights
        between a vector dimension and a cue with
        ``weights.loc[{'outcome_vector_dimensions': outcome_vector_dimension,
        'cue_vector_dimensions': cue_vector_dimension}]`` or
        ``weights.loc[vector_dimension].loc[cue_vector_dimension]``.

    """
    if cue_vectors is None and outcome_vectors is None:
        lambda_ = 1.0
        alpha = 1.0
        betas = (eta, eta)
        return ndl.ndl(events, alpha, betas, lambda_,
                       method=method, weights=weights, n_jobs=n_jobs,
                       n_outcomes_per_job=n_outcomes_per_job,
                       remove_duplicates=remove_duplicates, verbose=verbose,
                       temporary_directory=temporary_directory,
                       events_per_temporary_file=events_per_temporary_file)
    elif cue_vectors is not None and outcome_vectors is not None:
        return _wh_real_to_real(events, eta, cue_vectors, outcome_vectors,
                                method=method, weights=weights, n_jobs=n_jobs,
                                n_outcomes_per_job=n_outcomes_per_job,
                                remove_duplicates=remove_duplicates, verbose=verbose,
                                temporary_directory=temporary_directory,
                                events_per_temporary_file=events_per_temporary_file)
    elif cue_vectors is not None and outcome_vectors is None:
        lambda_ = 1.0
        betas = (eta, eta)
        return _wh_real_to_binary(events, betas, lambda_, cue_vectors,
                                  method=method, weights=weights, n_jobs=n_jobs,
                                  n_outcomes_per_job=n_outcomes_per_job,
                                  remove_duplicates=remove_duplicates, verbose=verbose,
                                  temporary_directory=temporary_directory,
                                  events_per_temporary_file=events_per_temporary_file)
    elif cue_vectors is None and outcome_vectors is not None:
        return _wh_binary_to_real(events, eta, outcome_vectors,
                                  method=method, weights=weights, n_jobs=n_jobs,
                                  n_outcomes_per_job=n_outcomes_per_job,
                                  remove_duplicates=remove_duplicates, verbose=verbose,
                                  temporary_directory=temporary_directory,
                                  events_per_temporary_file=events_per_temporary_file)
    # The if statements above are covering all cases.


def dict_wh(events, eta, cue_vectors, outcome_vectors, *,
            weights=None, inplace=False, remove_duplicates=None,
            make_data_array=False, verbose=False):
    """
    Calculate the weights for all_outcomes over all events in events.

    This is a pure python implementation using dicts.

    Notes
    -----
    The metadata will only be stored when `make_data_array` is True and then
    `dict_ndl` cannot be used to continue learning. At the moment there is no
    proper way to automatically store the meta data into the default dict.

    Parameters
    ----------
    events : generator or str
        generates cues, outcomes pairs or the path to the event file
    eta : float
        learning rate
    cue_vectors : xarray.DataArray
        matrix that contains the cue vectors for each cue
    outcome_vectors : xarray.DataArray
        matrix that contains the target vectors for each outcome
    weights : dict of dicts or xarray.DataArray or None
        initial weights
    inplace: {True, False}
        if True calculates the weightmatrix inplace
        if False creates a new weightmatrix to learn on
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    make_data_array : {False, True}
        if True makes a xarray.DataArray out of the dict of dicts.
    verbose : bool
        print some output if True.

    Returns
    -------
    weights : dict of dicts of floats
        the first dict has outcomes as keys and dicts as values
        the second dict has cues as keys and weights as values
        weights[outcome][cue] gives the weight between outcome and cue.

    or

    weights : xarray.DataArray
        with dimensions 'outcomes' and 'cues'. You can lookup the weights
        between a cue and an outcome with ``weights.loc[{'outcomes': outcome,
        'cues': cue}]`` or ``weights.loc[outcome].loc[cue]``.

    """

    if not isinstance(make_data_array, bool):
        raise ValueError("make_data_array must be True or False")

    if not (remove_duplicates is None or isinstance(remove_duplicates, bool)):
        raise ValueError("remove_duplicates must be None, True or False")

    wall_time_start = time.perf_counter()
    cpu_time_start = time.process_time()
    if isinstance(events, str):
        event_path = events
    else:
        event_path = ""
    attrs_to_update = None

    # weights can be seen as an infinite outcome by cue matrix
    # weights[outcome][cue]
    if weights is None:
        weights = WeightDict()
    elif isinstance(weights, WeightDict):
        attrs_to_update = weights.attrs
    elif isinstance(weights, xr.DataArray):
        raise NotImplementedError('initilizing with a xr.DataArray is not supported yet.')
        # weights_ini = weights
        # attrs_to_update = weights_ini.attrs
        # coords = weights_ini.coords
        # weights = WeightDict()
        # for outcome_index, outcome in enumerate(coords['outcomes'].values):
        #     for cue_index, cue in enumerate(coords['cues'].values):
        #         weights[outcome][cue] = weights_ini.item((outcome_index, cue_index))
    elif not isinstance(weights, defaultdict):
        raise ValueError('weights needs to be either defaultdict or None')

    if not inplace:
        weights = copy.deepcopy(weights)

    if isinstance(events, str):
        events = io.events_from_file(events)
    number_events = 0

    for cues, outcomes in events:
        number_events += 1
        if verbose and number_events % 1000:
            print('.', end='')
            sys.stdout.flush()
        if remove_duplicates is None:
            if (len(cues) != len(set(cues)) or
                    len(outcomes) != len(set(outcomes))):
                raise ValueError('cues or outcomes needs to be unique: cues '
                                 '"%s"; outcomes "%s"; use '
                                 'remove_duplicates=True' %
                                 (' '.join(cues), ' '.join(outcomes)))
        elif remove_duplicates:
            cues = set(cues)
            outcomes = set(outcomes)
        else:
            pass

        assert len(outcomes) == 1, 'for real_wh only one outcome is allowed per event'
        assert len(cues) == 1, 'for real_wh only one cue is allowed per event'

        cue_vec = cue_vectors.loc[cues[0]]
        outcome_vec = outcome_vectors.loc[outcomes[0]]

        for outcome_index in outcome_vec.coords['outcome_vector_dimensions']:
            outcome_index = str(outcome_index.values)
            prediction_strength = 0
            for cue_index in cue_vec.coords['cue_vector_dimensions']:
                cue_index = str(cue_index.values)
                prediction_strength += float(cue_vec.loc[cue_index] * weights[outcome_index][cue_index])
            error = float(outcome_vec.loc[outcome_index] - prediction_strength)

            for cue_index in cue_vec.coords['cue_vector_dimensions']:
                cue_index = str(cue_index.values)
                weights[outcome_index][cue_index] += float(eta * error * cue_vec.loc[cue_index])

    cpu_time_stop = time.process_time()
    wall_time_stop = time.perf_counter()
    cpu_time = cpu_time_stop - cpu_time_start
    wall_time = wall_time_stop - wall_time_start
    attrs = _attributes(event_path, number_events, eta, cpu_time, wall_time,
                        __name__ + "." + dict_wh.__name__, attrs=attrs_to_update)

    if make_data_array:
        outcome_dims = list(weights.keys())
        cue_dims = set()
        for outcome_dim in outcome_dims:
            cue_dims.update(set(weights[outcome_dim].keys()))

        cue_dims = list(cue_dims)

        weights_dict = weights
        shape = (len(outcome_dims), len(cue_dims))
        weights = xr.DataArray(np.zeros(shape), attrs=attrs,
                               coords={'outcome_vector_dimensions': outcome_dims,
                                       'cue_vector_dimensions': cue_dims},
                               dims=('outcome_vector_dimensions', 'cue_vector_dimensions'))

        for outcome_dim in outcome_dims:
            for cue_dim in cue_dims:
                weights.loc[{"outcome_vector_dimensions": outcome_dim,
                             "cue_vector_dimensions": cue_dim}] = weights_dict[outcome_dim][cue_dim]
    else:
        weights.attrs = attrs

    return weights


def _wh_binary_to_real(events, eta, outcome_vectors, *,
                       method='openmp', weights=None,
                       n_jobs=8, n_outcomes_per_job=10, remove_duplicates=None,
                       verbose=False, temporary_directory=None,
                       events_per_temporary_file=10000000):
    """
    Calculate the weights for all events using the Widrow-Hoff learning rule
    and training as outcomes on sematic vectors in semantics.

    This is a parallel python implementation using numpy, multithreading and
    the binary format defined in preprocess.py.

    Parameters
    ----------
    events : str
        path to the event file
    eta : float
        learning rate
    outcome_vectors : xarray.DataArray
        matrix that contains the target vectors for each outcome

    method : {'openmp', 'threading'}
    weights : None or xarray.DataArray
        the xarray.DataArray needs to have the dimensions 'cues' and 'outcomes'
    n_jobs : int
        an integer giving the number of threads in which the job should be
        executed
    n_outcomes_per_job : int
        an integer giving the number of outcomes that are processed in one job
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    verbose : bool
        print some output if True.
    temporary_directory : str
        path to directory to use for storing temporary files created;
        if none is provided, the operating system's default will
        be used (/tmp on unix)
    events_per_temporary_file: int
        Number of events in each temporary binary file. Has to be larger than 1

    Returns
    -------
    weights : xarray.DataArray
        with dimensions 'vector dimensions' and 'cues'. You can lookup the weights
        between a vector dimension and a cue with ``weights.loc[{'outcome_vector_dimensions': outcome_vector_dimension,
        'cues': cue}]`` or ``weights.loc[outcome_vector_dimension].loc[cue]``.

    """

    if not (remove_duplicates is None or isinstance(remove_duplicates, bool)):
        raise ValueError("remove_duplicates must be None, True or False")
    if not isinstance(events, str):
        raise ValueError("'events' need to be the path to a gzipped event file not {}".format(type(events)))

    weights_ini = weights
    wall_time_start = time.perf_counter()
    cpu_time_start = time.process_time()

    if type(outcome_vectors) == dict:
        # TODO: convert dict to xarray here
        raise NotImplementedError('dicts are not supported yet.')

    if not outcome_vectors.data.data.c_contiguous:
        raise ValueError('outcome_vectors have to be c_contiguous')

    # preprocessing
    n_events, cues, outcomes_from_events = count.cues_outcomes(events,
                                                               n_jobs=n_jobs,
                                                               verbose=verbose)
    cues = list(cues.keys())
    outcomes_from_events = list(outcomes_from_events.keys())
    outcomes = outcome_vectors.coords['outcomes'].values.tolist()
    cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues)))
    outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(outcomes)))

    # check for unseen outcomes in events
    if set(outcomes_from_events) - set(outcomes):
        raise ValueError("all outcomes in events need to be specified as rows in outcome_vectors")

    if weights is not None and weights.shape[0] != outcome_vectors.shape[1]:
        raise ValueError("outcome dimensions in weights need to match dimensions in outcome_vectors")

    shape = (outcome_vectors.shape[1], len(cue_map))

    # initialize weights
    if weights is None:
        weights = np.ascontiguousarray(np.zeros(shape, dtype=np.float64, order='C'))
    elif isinstance(weights, xr.DataArray):
        outcome_vector_dimensions = outcome_vectors.coords["outcome_vector_dimensions"].values.tolist()
        if not outcome_vector_dimensions == weights.coords["outcome_vector_dimensions"].values.tolist():
            raise ValueError('Outcome vector dimensions in weights and outcome_vectors do not match!')

        old_cues = weights.coords["cues"].values.tolist()
        new_cues = list(set(cues) - set(old_cues))
        cues = old_cues + new_cues
        cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues)))

        weights_tmp = np.concatenate((weights.values,
                                      np.zeros((len(outcome_vector_dimensions), len(new_cues)),
                                               dtype=np.float64, order='C')),
                                     axis=1)
        weights = np.ascontiguousarray(weights_tmp)

        del weights_tmp, old_cues, new_cues, outcome_vector_dimensions
    else:
        raise ValueError('weights need to be None or xarray.DataArray with method=%s' % method)

    with tempfile.TemporaryDirectory(prefix="pyndl", dir=temporary_directory) as binary_path:
        number_events = preprocess.create_binary_event_files(events, binary_path, cue_map,
                                                             outcome_map, overwrite=True,
                                                             n_jobs=n_jobs,
                                                             events_per_file=events_per_temporary_file,
                                                             remove_duplicates=remove_duplicates,
                                                             verbose=verbose)
        assert n_events == number_events, (str(n_events) + ' ' + str(number_events))
        binary_files = [os.path.join(binary_path, binary_file)
                        for binary_file in os.listdir(binary_path)
                        if os.path.isfile(os.path.join(binary_path, binary_file))]
        # sort binary files as they were created
        binary_files.sort(key=lambda filename: int(os.path.basename(filename)[9:-4]))
        if verbose:
            print('start learning...')
        # learning
        if not weights.data.c_contiguous:
            raise ValueError('weights has to be c_contiguous')
        if method == 'openmp':
            if not sys.platform.startswith('linux'):
                raise NotImplementedError("OpenMP is linux only at the moment."
                                          "Use method='threading' instead.")
            ndl_openmp.learn_inplace_binary_to_real(binary_files,
                                                    eta,
                                                    outcome_vectors.data,
                                                    weights,
                                                    n_outcomes_per_job,
                                                    n_jobs)
        # elif method == 'threading':
        #    part_lists = ndl.slice_list(all_outcome_indices, n_outcomes_per_job)

        #    working_queue = Queue(len(part_lists))
        #    threads = []
        #    queue_lock = threading.Lock()

        #    def worker():
        #        while True:
        #            with queue_lock:
        #                if working_queue.empty():
        #                    break
        #                data = working_queue.get()
        #            ndl_parallel.learn_inplace_2(binary_files, outcome_vectors,
        #                                         eta, weights, data)

        #    with queue_lock:
        #        for partlist in part_lists:
        #            working_queue.put(np.array(partlist, dtype=np.uint32))

        #    for _ in range(n_jobs):
        #        thread = threading.Thread(target=worker)
        #        thread.start()
        #        threads.append(thread)

        #    for thread in threads:
        #        thread.join()
        else:
            raise ValueError('method needs to be either "threading" or "openmp"')

    cpu_time_stop = time.process_time()
    wall_time_stop = time.perf_counter()
    cpu_time = cpu_time_stop - cpu_time_start
    wall_time = wall_time_stop - wall_time_start

    if weights_ini is not None:
        attrs_to_be_updated = weights_ini.attrs
    else:
        attrs_to_be_updated = None

    attrs = _attributes(events, number_events, eta, cpu_time, wall_time,
                        __name__ + "." + ndl.__name__, method=method, attrs=attrs_to_be_updated)

    # post-processing
    weights = xr.DataArray(weights, coords=[('outcome_vector_dimensions',
                                             outcome_vectors.coords['outcome_vector_dimensions'].data),
                                            ('cues', cues)], attrs=attrs)
    return weights


def _wh_real_to_binary(events, betas, lambda_, cue_vectors, *,
                       method='openmp', weights=None,
                       n_jobs=8, n_outcomes_per_job=10, remove_duplicates=None,
                       verbose=False, temporary_directory=None,
                       events_per_temporary_file=10000000):
    """
    Calculate the weights for all events using the Widrow-Hoff learning rule
    and training as cue_vectors on outcomes.

    This is a parallel python implementation using numpy, multithreading and
    the binary format defined in preprocess.py.

    Parameters
    ----------
    events : str
        path to the event file
    betas : (float, float)
        one value for successful prediction (reward) one for punishment
    lambda\\_ : float
    cue_vectors : xarray.DataArray
        matrix that contains the cue vectors for each cue
    method : {'openmp', 'threading'}
    weights : None or xarray.DataArray
        the xarray.DataArray needs to have the dimensions 'cues' and 'outcomes'
    n_jobs : int
        an integer giving the number of threads in which the job should be
        executed
    n_outcomes_per_job : int
        an integer giving the number of outcomes that are processed in one job
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    verbose : bool
        print some output if True.
    temporary_directory : str
        path to directory to use for storing temporary files created;
        if none is provided, the operating system's default will
        be used (/tmp on unix)
    events_per_temporary_file: int
        Number of events in each temporary binary file. Has to be larger than 1

    Returns
    -------
    weights : xarray.DataArray
        with dimensions outcomes and cue_vector_dimensions. You can lookup the
        weights between an outcome and a cue_vector_dimension with
        ``weights.loc[{'outcome': outcome, 'cue_vector_dimensions':
        cue_vector_dimension}]`` or
        ``weights.loc[outcome].loc[cue_vector_dimension]``.

    """
    if not (remove_duplicates is None or isinstance(remove_duplicates, bool)):
        raise ValueError("remove_duplicates must be None, True or False")

    if not isinstance(events, str):
        raise ValueError("'events' need to be the path to a gzipped event file not {}".format(type(events)))

    if type(cue_vectors) == dict:
        # TODO: convert dict to xarray here
        raise NotImplementedError('dicts are not supported yet.')

    if not cue_vectors.data.data.c_contiguous:
        raise ValueError('cue_vectors have to be c_contiguous')

    weights_ini = weights
    wall_time_start = time.perf_counter()
    cpu_time_start = time.process_time()

    # preprocessing
    n_events, cues_from_events, outcomes_from_events = count.cues_outcomes(events,
                                                                           n_jobs=n_jobs,
                                                                           verbose=verbose)

    cues_from_events = list(cues_from_events.keys())
    cues = list(cue_vectors.coords['cues'].data)
    outcomes = list(outcomes_from_events.keys())

    cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues)))
    outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(outcomes)))

    if set(cues_from_events) - set(cues):
        raise ValueError("all cues in events need to be specified as rows in cue_vectors")

    del outcomes_from_events, cues_from_events

    shape = (len(outcomes), cue_vectors.shape[1])

    if not cue_vectors.dims[1] == 'cue_vector_dimensions':
        raise ValueError("The second dimension of the 'cue_vectors' has to be named 'cue_vector_dimensions'.")

    cue_vector_dimensions = cue_vectors['cue_vector_dimensions']

    # initialize weights
    if weights is None:
        weights = np.ascontiguousarray(np.zeros(shape, dtype=np.float64, order='C'))
    elif isinstance(weights, xr.DataArray):
        if not all(cue_vector_dimensions == weights['cue_vector_dimensions']):
            raise ValueError("Cue vector dimensions names do not match in weights and cue_vectors")

        old_outcomes = weights.coords["outcomes"].values.tolist()
        new_outcomes = list(set(outcomes) - set(old_outcomes))
        outcomes = old_outcomes + new_outcomes

        outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(outcomes)))
        weights_tmp = np.concatenate((weights.values,
                                      np.zeros((len(new_outcomes), len(cue_vector_dimensions)),
                                               dtype=np.float64, order='C')),
                                     axis=0)
        weights = np.ascontiguousarray(weights_tmp)
        del weights_tmp, old_outcomes, new_outcomes
    else:
        raise ValueError('weights need to be None or xarray.DataArray with method=%s' % method)

    weights = xr.DataArray(weights,
                           coords={'outcomes': outcomes, 'cue_vector_dimensions': cue_vector_dimensions},
                           dims=('outcomes', 'cue_vector_dimensions'))
    del shape, cue_vector_dimensions

    with tempfile.TemporaryDirectory(prefix="pyndl", dir=temporary_directory) as binary_path:
        number_events = preprocess.create_binary_event_files(events, binary_path, cue_map,
                                                             outcome_map, overwrite=True,
                                                             n_jobs=n_jobs,
                                                             events_per_file=events_per_temporary_file,
                                                             remove_duplicates=remove_duplicates,
                                                             verbose=verbose)
        assert n_events == number_events, (str(n_events) + ' ' + str(number_events))
        binary_files = [os.path.join(binary_path, binary_file)
                        for binary_file in os.listdir(binary_path)
                        if os.path.isfile(os.path.join(binary_path, binary_file))]
        # sort binary files as they were created
        binary_files.sort(key=lambda filename: int(os.path.basename(filename)[9:-4]))
        if verbose:
            print('start learning...')
        # learning
        if not weights.data.data.c_contiguous:
            raise ValueError('weights has to be c_contiguous')
        if method == 'openmp':
            if not sys.platform.startswith('linux'):
                raise NotImplementedError("OpenMP is linux only at the moment."
                                          "Use method='threading' instead.")
            beta1, beta2 = betas
            ndl_openmp.learn_inplace_real_to_binary(binary_files,
                                                    beta1,
                                                    beta2,
                                                    lambda_,
                                                    cue_vectors.data,
                                                    weights.data,
                                                    n_outcomes_per_job,
                                                    n_jobs)

        weights = weights.reset_coords(drop=True)

    cpu_time_stop = time.process_time()
    wall_time_stop = time.perf_counter()
    cpu_time = cpu_time_stop - cpu_time_start
    wall_time = wall_time_stop - wall_time_start

    if weights_ini is not None:
        attrs_to_be_updated = weights_ini.attrs
    else:
        attrs_to_be_updated = None

    attrs = ndl._attributes(events, number_events, 'cue_vectors', betas, lambda_, cpu_time, wall_time,
                            __name__ + "." + ndl.__name__, method=method, attrs=attrs_to_be_updated)

    # post-processing
    weights.attrs = attrs
    return weights


def _wh_real_to_real(events, eta, cue_vectors, outcome_vectors, *,
                     method='openmp', weights=None,
                     n_jobs=8, n_outcomes_per_job=10, remove_duplicates=None,
                     verbose=False, temporary_directory=None,
                     events_per_temporary_file=10000000):
    """
    Calculate the weights for all events using the Widrow-Hoff learning rule
    and training as outcomes on sematic vectors in semantics.

    This is a parallel python implementation using numpy, multithreading and
    the binary format defined in preprocess.py.

    Parameters
    ----------
    events : str
        path to the event file
    eta : float
        learning rate
    cue_vectors : xarray.DataArray
        matrix that contains the cue vectors for each cue
    outcome_vectors : xarray.DataArray
        matrix that contains the target vectors for each outcome
    method : {'openmp', 'threading', 'numpy'}
    weights : None or xarray.DataArray
        the xarray.DataArray needs to have the dimensions 'cues' and 'outcomes'
    n_jobs : int
        an integer giving the number of threads in which the job should be
        executed
    n_outcomes_per_job : int
        an integer giving the number of outcomes that are processed in one job
    remove_duplicates : {None, True, False}
        if None though a ValueError when the same cue is present multiple times
        in the same event; True make cues and outcomes unique per event; False
        keep multiple instances of the same cue or outcome (this is usually not
        preferred!)
    verbose : bool
        print some output if True.
    temporary_directory : str
        path to directory to use for storing temporary files created;
        if none is provided, the operating system's default will
        be used (/tmp on unix)
    events_per_temporary_file: int
        Number of events in each temporary binary file. Has to be larger than 1

    Returns
    -------
    weights : xarray.DataArray
        with dimensions 'outcome_vector_dimensions' and
        'cue_vector_dimensions'. You can lookup the weights between a
        outcome_vector_dimension and a cue_vector_dimension with
        ``weights.loc[{'outcome_vector_dimensions': outcome_vector_dimension,
            'cue_vector_dimensions': cue_vector_dimension}]`` or
        ``weights.loc[outcome_vector_dimension].loc[cue_vector_dimension]``.

    """

    if not (remove_duplicates is None or isinstance(remove_duplicates, bool)):
        raise ValueError("remove_duplicates must be None, True or False")
    if not isinstance(events, str):
        raise ValueError("'events' need to be the path to a gzipped event file not {}".format(type(events)))

    weights_ini = weights
    wall_time_start = time.perf_counter()
    cpu_time_start = time.process_time()

    if type(cue_vectors) == dict:
        # TODO: convert dict to xarray here
        raise NotImplementedError('dicts are not supported yet.')

    if type(outcome_vectors) == dict:
        # TODO: convert dict to xarray here
        raise NotImplementedError('dicts are not supported yet.')

    if not cue_vectors.data.data.c_contiguous:
        raise ValueError('cue_vectors have to be c_contiguous')

    if not outcome_vectors.data.data.c_contiguous:
        raise ValueError('outcome_vectors have to be c_contiguous')

    # preprocessing
    n_events, cues_from_events, outcomes_from_events = count.cues_outcomes(events,
                                                                           n_jobs=n_jobs,
                                                                           verbose=verbose)

    cues_from_events = list(cues_from_events.keys())
    cues = list(cue_vectors.coords['cues'].data)
    outcomes_from_events = list(outcomes_from_events.keys())
    outcomes = list(outcome_vectors.coords['outcomes'].data)
    cue_map = OrderedDict(((cue, ii) for ii, cue in enumerate(cues)))
    outcome_map = OrderedDict(((outcome, ii) for ii, outcome in enumerate(outcomes)))

    # check for unseen outcomes in events
    if set(outcomes_from_events) - set(outcomes):
        raise ValueError("all outcomes in events need to be specified as rows in outcome_vectors")
    if set(cues_from_events) - set(cues):
        raise ValueError("all cues in events need to be specified as rows in cue_vectors")

    del outcomes_from_events, cues_from_events

    shape = (outcome_vectors.shape[1], cue_vectors.shape[1])

    if not outcome_vectors.dims[1] == 'outcome_vector_dimensions':
        raise ValueError("The second dimension of the 'outcome_vectors' has to be named 'outcome_vector_dimensions'.")
    if not cue_vectors.dims[1] == 'cue_vector_dimensions':
        raise ValueError("The second dimension of the 'cue_vectors' has to be named 'cue_vector_dimensions'.")

    cue_dims = cue_vectors['cue_vector_dimensions']
    outcome_dims = outcome_vectors['outcome_vector_dimensions']

    # initialize weights
    if weights is None:
        weights = np.ascontiguousarray(np.zeros(shape, dtype=np.float64, order='C'))
        weights = xr.DataArray(weights,
                               coords={'outcome_vector_dimensions': outcome_dims, 'cue_vector_dimensions': cue_dims},
                               dims=('outcome_vector_dimensions', 'cue_vector_dimensions'))
        del cue_dims, outcome_dims
    elif isinstance(weights, xr.DataArray):
        if not weights.shape == shape:
            raise ValueError("Vector dimensions need to match in continued learning!")
        if not all(outcome_dims == weights['outcome_vector_dimensions']):
            raise ValueError("Outcome vector dimensions names do not match in weights and outcome_vectors")
        if not all(cue_dims == weights['cue_vector_dimensions']):
            raise ValueError("Cue vector dimensions names do not match in weights and cue_vectors")
        # align the cue and outcome vector dimension names between the old weights
        # and the cue_vectors / outcome_vectors
        weights = weights.loc[{'outcome_vector_dimensions': outcome_dims, 'cue_vector_dimensions': cue_dims}]
        weights = weights.copy()
    else:
        raise ValueError('weights need to be None or xarray.DataArray with method=%s' % method)
    del shape

    if not weights.data.data.c_contiguous:
        raise ValueError('weights has to be c_contiguous')
    if method == 'numpy':
        event_generator = io.events_from_file(events)
        number_events = 0

        if verbose:
            print('start learning...')
        for cues, outcomes in event_generator:
            number_events += 1
            if verbose and number_events % 1000:
                print('.', end='')
                sys.stdout.flush()
            if remove_duplicates is None:
                if (len(cues) != len(set(cues)) or
                        len(outcomes) != len(set(outcomes))):
                    raise ValueError('cues or outcomes needs to be unique: cues '
                                     '"%s"; outcomes "%s"; use '
                                     'remove_duplicates=True' %
                                     (' '.join(cues), ' '.join(outcomes)))
            elif remove_duplicates:
                cues = set(cues)
                outcomes = set(outcomes)
            else:
                pass

            # TODO: implement multiple cues / outcomes in numpy
            assert len(outcomes) == 1, 'for method=numpy only one outcome is allowed per event'
            assert len(cues) == 1, 'for method_numpy only one cue is allowed per event'

            cue_vec = cue_vectors.loc[cues[0]]
            outcome_vec = outcome_vectors.loc[outcomes[0]]

            prediction_vec = weights.dot(cue_vec)  # why is the @ not working?
            error = outcome_vec - prediction_vec
            weights += eta * error * cue_vec  # broadcasted array multiplication
            # NOTE: we could calculate the same weights first on the first half
            # of the outcome vector dimensions and then on the second half and
            # row bind both in the end. Do we? Yes, and we can use this to
            # multiprocess the numpy computation.
    elif method in ('openmp', 'threading'):
        with tempfile.TemporaryDirectory(prefix="pyndl", dir=temporary_directory) as binary_path:
            number_events = preprocess.create_binary_event_files(events, binary_path, cue_map,
                                                                 outcome_map, overwrite=True,
                                                                 n_jobs=n_jobs,
                                                                 events_per_file=events_per_temporary_file,
                                                                 remove_duplicates=remove_duplicates,
                                                                 verbose=verbose)
            assert n_events == number_events, (str(n_events) + ' ' + str(number_events))
            binary_files = [os.path.join(binary_path, binary_file)
                            for binary_file in os.listdir(binary_path)
                            if os.path.isfile(os.path.join(binary_path, binary_file))]
            # sort binary files as they were created
            binary_files.sort(key=lambda filename: int(os.path.basename(filename)[9:-4]))
            if verbose:
                print('start learning...')
            # learning
            if method == 'openmp':
                if not sys.platform.startswith('linux'):
                    raise NotImplementedError("OpenMP is linux only at the moment."
                                              "Use method='threading' instead.")
                ndl_openmp.learn_inplace_real_to_real(binary_files,
                                                      eta,
                                                      cue_vectors.data,
                                                      outcome_vectors.data,
                                                      weights.data,
                                                      n_outcomes_per_job,
                                                      n_jobs)
            else:
                # TODO: implement threading
                raise ValueError('TODO: for now: method needs to be "numpy" or "openmp"')

            weights = weights.reset_coords(drop=True)
    else:
        raise ValueError('method needs to be "numpy" or "openmp"')

    cpu_time_stop = time.process_time()
    wall_time_stop = time.perf_counter()
    cpu_time = cpu_time_stop - cpu_time_start
    wall_time = wall_time_stop - wall_time_start

    if weights_ini is not None:
        attrs_to_be_updated = weights_ini.attrs
    else:
        attrs_to_be_updated = None

    attrs = _attributes(events, number_events, eta, cpu_time, wall_time,
                        __name__ + "." + ndl.__name__, method=method, attrs=attrs_to_be_updated)

    # post-processing
    weights.attrs = attrs
    return weights


def _attributes(event_path, number_events, eta, cpu_time,
                wall_time, function, method=None, attrs=None):

    width = max([len(ss) for ss in (event_path,
                                    str(number_events),
                                    str(eta),
                                    function,
                                    str(method),
                                    socket.gethostname(),
                                    getpass.getuser())])
    width = max(19, width)

    def _format(value):
        return '{0: <{width}}'.format(value, width=width)

    new_attrs = {'date': _format(time.strftime("%Y-%m-%d %H:%M:%S")),
                 'event_path': _format(event_path),
                 'number_events': _format(number_events),
                 'lambda': _format(str(eta)),
                 'function': _format(function),
                 'method': _format(str(method)),
                 'cpu_time': _format(str(cpu_time)),
                 'wall_time': _format(str(wall_time)),
                 'hostname': _format(socket.gethostname()),
                 'username': _format(getpass.getuser()),
                 'pyndl': _format(pyndl_version),
                 'numpy': _format(np.__version__),
                 'pandas': _format(pd.__version__),
                 'xarray': _format(xr.__version__),
                 'cython': _format(cython.__version__)}

    if attrs is not None:
        for key in set(attrs.keys()) | set(new_attrs.keys()):
            if key in attrs:
                old_val = attrs[key]
            else:
                old_val = ''
            if key in new_attrs:
                new_val = new_attrs[key]
            else:
                new_val = ''
            new_attrs[key] = old_val + ' | ' + new_val
    return new_attrs
